import asyncio
import socket
from abc import ABC, abstractmethod
from typing import Tuple, List, Set
from xml.etree import ElementTree

import aiohttp
import async_timeout
from aiohttp import ClientResponse

from .data import IrcelineFeature

_rio_wfs_base_url = 'https://geo.irceline.be/wfs'
_forecast_wms_base_url = 'https://geo.irceline.be/forecast/wms'
_rio_ifdm_wms_base_url = 'https://geobelair.irceline.be/rioifdm/wms'
_user_agent = 'github.com/jdejaegh/python-irceline'


class IrcelineApiError(Exception):
    """Exception to indicate an API error."""


class IrcelineBaseClient(ABC):
    def __init__(self, session: aiohttp.ClientSession) -> None:
        self._session = session

    @abstractmethod
    async def get_data(self,
                       features: List[IrcelineFeature],
                       position: Tuple[float, float]) -> dict:
        pass

    @abstractmethod
    def get_capabilities(self) -> Set[str]:
        pass

    async def _api_wrapper(self, url: str, querystring: dict = None, headers: dict = None, method: str = 'GET'):
        """
        Call the URL with the specified query string. Raises exception for >= 400 response code
        :param url: base URL
        :param querystring: dict to build the query string
        :return: response from the client
        """
        if headers is None:
            headers = dict()
        if 'User-Agent' not in headers:
            headers |= {'User-Agent': _user_agent}
        try:
            async with async_timeout.timeout(60):
                response = await self._session.request(
                    method=method,
                    url=url,
                    params=querystring,
                    headers=headers
                )
                response.raise_for_status()
                return response

        except asyncio.TimeoutError as exception:
            raise IrcelineApiError("Timeout error fetching information") from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            raise IrcelineApiError("Error fetching information") from exception
        except Exception as exception:  # pylint: disable=broad-except
            raise IrcelineApiError(f"Something really wrong happened! {exception}") from exception


class IrcelineBaseWmsClient(IrcelineBaseClient, ABC):
    _default_querystring = {"service": "WMS",
                            "version": "1.1.1",
                            "request": "GetFeatureInfo",
                            "info_format": "application/json",
                            "width": "1",
                            "height": "1",
                            "srs": "EPSG:4326",
                            "X": "1",
                            "Y": "1"}
    _epsilon = 0.00001
    _base_url = None

    @staticmethod
    def _parse_capabilities(xml_string: str) -> Set[str]:
        try:
            root = ElementTree.fromstring(xml_string)
        except ElementTree.ParseError:
            return set()

        path = './/Capability/Layer/Layer/Name'
        feature_type_names = {t.text for t in root.findall(path)}
        return feature_type_names

    async def get_capabilities(self) -> Set[str]:
        """
        Fetch the list of possible features from the WMS server
        :return: set of features available on the WMS server
        """
        querystring = {"service": "WMS",
                       "version": "1.1.1",
                       "request": "GetCapabilities"}
        r: ClientResponse = await self._api_wrapper(self._base_url, querystring)

        return self._parse_capabilities(await r.text())



from .api import IrcelineApiError
from .data import RioFeature, ForecastFeature, FeatureValue, RioIfdmFeature
from .forecast import IrcelineForecastClient
from .rio import IrcelineRioClient, IrcelineRioIfdmClient

__version__ = '3.0.2'



import asyncio
from datetime import date, timedelta, datetime
from itertools import product
from typing import List, Tuple, Dict

from aiohttp import ClientResponse, ClientResponseError

from .api import IrcelineApiError, IrcelineBaseWmsClient, _forecast_wms_base_url
from .data import ForecastFeature, FeatureValue


class IrcelineForecastClient(IrcelineBaseWmsClient):
    _base_url = _forecast_wms_base_url

    async def get_data(self,
                       features: List[ForecastFeature],
                       position: Tuple[float, float]
                       ) -> Dict[Tuple[ForecastFeature, date], FeatureValue]:
        """
        Get forecasted concentrations for the given features at the given position. The forecasts are downloaded for
        the specified day and the 3 next days as well
        :param features: pollutants to get the forecasts for
        :param position: (lat, long)
        :return: dict where key is (ForecastFeature, date of the forecast) and value is a FeatureValue
        """
        timestamp = date.today()
        result = dict()
        lat, lon = position
        base_querystring = (self._default_querystring |
                            {"bbox": f"{lon},{lat},{lon + self._epsilon},{lat + self._epsilon}"})

        tasks = [asyncio.create_task(self._get_single_feature(base_querystring, d, feature, timestamp))
                 for feature, d in product(features, range(4))]
        results = await asyncio.gather(*tasks)

        for r in results:
            result |= r

        return result

    async def _get_single_feature(self, base_querystring: dict, d: int, feature: ForecastFeature,
                                  timestamp: date) -> dict:
        result = dict()

        querystring = base_querystring | {"layers": f"{feature}_d{d}",
                                          "query_layers": f"{feature}_d{d}"}
        try:
            r: ClientResponse = await self._api_wrapper(self._base_url, querystring)
            r: dict = await r.json()
            result[(feature, timestamp + timedelta(days=d))] = FeatureValue(
                value=r.get('features', [{}])[0].get('properties', {}).get('GRAY_INDEX'),
                timestamp=datetime.fromisoformat(r.get('timeStamp')) if 'timeStamp' in r else None)
        except (IrcelineApiError, ClientResponseError, IndexError):
            result[(feature, timestamp + timedelta(days=d))] = FeatureValue(value=None, timestamp=None)
        return result



from datetime import datetime
from itertools import product
from unittest.mock import call

from freezegun import freeze_time

from src.open_irceline import IrcelineForecastClient, ForecastFeature, FeatureValue
from src.open_irceline.api import _user_agent, _forecast_wms_base_url
from tests.conftest import get_api_data, get_mock_session


def test_parse_capabilities():
    data = get_api_data('forecast_wms_capabilities.xml', plain=True)
    result = IrcelineForecastClient._parse_capabilities(data)

    expected = {'o3_maxhmean_wl_d3', 'pm25_dmean_wl_d0', 'o3_max8hmean_chimv2022_d2', 'no2_maxhmean_tf_d2',
                'belaqi_forecast_chimv2022_d2', 'pm25_dmean_chimv2022_d3', 'pm10_dmean_chimv2022_d0',
                'no2_maxhmean_wl_d0', 'no2_maxhmean_d2', 'no2_dmean_chimv2022_d2', 'o3_maxhmean_chimv2022_d3',
                'pm25_dmean_wl_d3', 'o3_maxhmean_chimv2022_d0', 'pm25_dmean', 'pm25_dmean_tf_d0', 'no2_dmean_wl_d2',
                'o3_max8hmean_chimv2022_d3', 'pm25_dmean_d2', 'o3_max8hmean_chimv2022_d0', 'o3_maxhmean_wl_d2',
                'no2_maxhmean_wl_d1', 'pm10_dmean_tf_d2', 'pm25_dmean_d1', 'o3_maxhmean_chimv2022_d2',
                'pm10_dmean_chimv2022_d2', 'o3_maxhmean_vl', 'belaqi_wl_d2', 'pm10_dmean_wl', 'pm10_dmean_d2',
                'no2_dmean_wl_d0', 'no2_dmean_d1', 'o3_maxhmean_d2', 'o3_maxhmean_wl', 'pm25_dmean_wl_d2',
                'o3_maxhmean_d3', 'o3_max8hmean_wl_d3', 'belaqi_d0', 'no2_maxhmean_wl_d2', 'no2_maxhmean_wl',
                'pm10_dmean_wl_d1', 'no2_dmean_chimv2022_d3', 'o3_maxhmean_tf_d1', 'pm25_dmean_vl', 'pm10_dmean_d0',
                'o3_max8hmean_d0', 'o3_max8hmean_d2', 'no2_maxhmean_vl', 'o3_max8hmean_chimv2022_d1', 'pm10_dmean',
                'pm10_dmean_wl_d2', 'euaqi_d3', 'belaqi_d1', 'o3_max8hmean_d1', 'o3_maxhmean_chimv2022_d1', 'belaqi_vl',
                'belaqi_wl_d0', 'no2_dmean_chimv2022_d0', 'pm25_dmean_wl_d1', 'pm25_dmean_tf_d2', 'no2_dmean_d2',
                'o3_maxhmean', 'belaqi_wl', 'no2_maxhmean_d0', 'no2_maxhmean_d3', 'o3_max8hmean_d3', 'euaqi_forecast',
                'o3_max8hmean_wl_d1', 'pm10_dmean_chimv2022_d3', 'no2_maxhmean_wl_d3', 'o3_maxhmean_d1',
                'no2_dmean_wl_d1', 'o3_maxhmean_wl_d1', 'no2_dmean_d3', 'belaqi_d3', 'belaqi', 'pm25_dmean_d3',
                'belaqi_forecast', 'no2_dmean_d0', 'pm25_dmean_chimv2022_d1', 'belaqi_wl_d1', 'pm10_dmean_d3',
                'no2_dmean_wl_d3', 'pm25_dmean_tf_d1', 'euaqi_d0', 'o3_maxhmean_wl_d0', 'belaqi_forecast_chimv2022_d3',
                'no2_dmean_chimv2022_d1', 'o3_max8hmean_wl_d0', 'o3_max8hmean_wl_d2', 'pm10_dmean_chimv2022_d1',
                'pm10_dmean_wl_d3', 'pm25_dmean_wl', 'belaqi_forecast_chimv2022_d1', 'euaqi_d2', 'pm10_dmean_d1',
                'belaqi_wl_d3', 'belaqi_forecast_chimv2022_d0', 'o3_maxhmean_tf_d0', 'euaqi_d1', 'no2_maxhmean',
                'pm25_dmean_chimv2022_d2', 'belaqi_d2', 'pm25_dmean_d0', 'no2_maxhmean_tf_d0', 'pm10_dmean_tf_d0',
                'pm25_dmean_chimv2022_d0', 'o3_maxhmean_d0', 'pm10_dmean_tf_d1', 'pm10_dmean_vl', 'no2_maxhmean_tf_d1',
                'o3_maxhmean_tf_d2', 'pm10_dmean_wl_d0', 'no2_maxhmean_d1'}

    assert result == expected

    for f, d in product(ForecastFeature, range(4)):
        assert f"{f.split(':')[1]}_d{d}" in result


async def test_aget_capabilities():
    session = get_mock_session(text_file='forecast_wms_capabilities.xml')

    client = IrcelineForecastClient(session)
    _ = await client.get_capabilities()

    session.request.assert_called_once_with(
        method='GET',
        url=_forecast_wms_base_url,
        params={"service": "WMS",
                "version": "1.1.1",
                "request": "GetCapabilities"},
        headers={'User-Agent': _user_agent}
    )


@freeze_time(datetime.fromisoformat("2024-06-30T13:00:21.520Z"))
async def test_api_forecast_error():
    pos = (50.4657, 4.8647)
    session = get_mock_session('forecast_wms_feature_info_invalid.json')

    client = IrcelineForecastClient(session)

    features = [ForecastFeature.NO2_DMEAN, ForecastFeature.O3_MAXHMEAN]
    result = await client.get_data(features, pos)

    for k, v in result.items():
        assert v == FeatureValue(timestamp=datetime.fromisoformat("2024-06-30T13:00:21.520Z"), value=None)


async def test_api_forecast_no_field():
    pos = (50.4657, 4.8647)
    session = get_mock_session('forecast_wms_feature_info_no_field.json')

    client = IrcelineForecastClient(session)

    features = [ForecastFeature.NO2_DMEAN, ForecastFeature.O3_MAXHMEAN]
    result = await client.get_data(features, pos)

    for k, v in result.items():
        assert v == FeatureValue(timestamp=None, value=None)


async def test_api_forecast():
    pos = (50.4657, 4.8647)
    lat, lon = pos
    session = get_mock_session('forecast_wms_feature_info.json')

    client = IrcelineForecastClient(session)

    features = [ForecastFeature.NO2_DMEAN, ForecastFeature.O3_MAXHMEAN]
    result = await client.get_data(features, pos)

    base = {"service": "WMS",
            "version": "1.1.1",
            "request": "GetFeatureInfo",
            "info_format": "application/json",
            "width": "1",
            "height": "1",
            "srs": "EPSG:4326",
            "bbox": f"{lon},{lat},{lon + 0.00001},{lat + 0.00001}",
            "X": "1",
            "Y": "1"}

    calls = [call(
        method='GET',
        url=_forecast_wms_base_url,
        params=base | {"layers": f"{feature}_d{d}",
                       "query_layers": f"{feature}_d{d}"},
        headers={'User-Agent': _user_agent},
    )
        for feature, d in product(features, range(4))]

    session.request.assert_has_calls(calls, any_order=True)

    for k, v in result.items():
        assert v['value'] == 10.853286743164062


def test_parse_capabilities_with_error():
    result = IrcelineForecastClient._parse_capabilities("wow there no valid XML")
    assert result == set()



import asyncio
from datetime import datetime, date, UTC, timedelta
from typing import List, Tuple, Dict, Set
from xml.etree import ElementTree

from aiohttp import ClientResponse, ClientResponseError

from .api import IrcelineBaseClient, _rio_wfs_base_url, IrcelineApiError, _rio_ifdm_wms_base_url, IrcelineBaseWmsClient
from .data import RioFeature, FeatureValue, RioIfdmFeature
from .utils import epsg_transform


class IrcelineRioClient(IrcelineBaseClient):
    """
    API client for RIO interpolated IRCEL - CELINE open data
    RIO is more coarse grained for interpolation than RIO IFDM and allows to request multiple features in the same
        request, which may be faster.
    """

    async def get_data(self,
                       features: List[RioFeature],
                       position: Tuple[float, float],
                       timestamp: datetime | date | None = None
                       ) -> Dict[RioFeature, FeatureValue]:
        """
        Call the WFS API to get the interpolated level of RioFeature. Raises exception upon API error
        :param timestamp: datetime for which to get the data for
        :param features: list of RioFeature to fetch from the API
        :param position: decimal degrees pair of coordinates
        :return: dict with the response (key is RioFeature, value is FeatureValue with actual value and timestamp)
        """
        if timestamp is None:
            timestamp = datetime.now(UTC)
        # Remove one hour/day from timestamp to handle case where the hour just passed but the data is not yet there
        # (e.g. 5.01 PM, but the most recent data is for 4.00 PM)
        if isinstance(timestamp, datetime):
            timestamp = timestamp.replace(microsecond=0, second=0, minute=0) - timedelta(hours=1)
            timestamp = timestamp.isoformat()
            key = 'timestamp'
        elif isinstance(timestamp, date):
            timestamp = timestamp - timedelta(days=1)
            timestamp = timestamp.isoformat()
            key = 'date'
        else:
            raise IrcelineApiError(f"Wrong parameter type for timestamp: {type(timestamp)}")

        lat, lon = epsg_transform(position)
        querystring = {"service": "WFS",
                       "version": "1.3.0",
                       "request": "GetFeature",
                       "outputFormat": "application/json",
                       "typeName": ",".join(features),
                       "cql_filter":
                           f"{key}>='{timestamp}'"
                           f" AND "
                           f"INTERSECTS(the_geom, POINT ({lat} {lon}))"}
        r: ClientResponse = await self._api_wrapper(_rio_wfs_base_url, querystring)
        return self._format_result('rio', await r.json(), features)

    async def get_capabilities(self) -> Set[str]:
        """
        Fetch the list of possible features from the WFS server
        :return: set of features available on the WFS server
        """
        querystring = {"service": "WFS",
                       "version": "1.3.0",
                       "request": "GetCapabilities"}
        r: ClientResponse = await self._api_wrapper(_rio_wfs_base_url, querystring)

        return self._parse_capabilities(await r.text())

    @staticmethod
    def _parse_capabilities(xml_string: str) -> Set[str]:
        """
        From an XML string obtained with GetCapabilities, generate a set of feature names
        :param xml_string: XML string to parse
        :return: set of FeatureType Names found in the XML document
        """
        try:
            root = ElementTree.fromstring(xml_string)
        except ElementTree.ParseError:
            return set()
        # noinspection HttpUrlsUsage
        # We never connect to the URL, it is just the namespace in the XML
        namespaces = {
            'wfs': 'http://www.opengis.net/wfs',
        }
        path = './/wfs:FeatureTypeList/wfs:FeatureType/wfs:Name'
        feature_type_names = {t.text for t in root.findall(path, namespaces)}
        return feature_type_names

    @staticmethod
    def _format_result(prefix: str, data: dict, features: List[RioFeature]) -> dict:
        """
        Format the JSON dict returned by the WFS service into a more practical dict to use with only the latest measure
        for each feature requested
        :param prefix: namespace of the feature (e.g. rio), without the colon
        :param data: JSON dict value as returned by the API
        :param features: RioFeatures wanted in the final dict
        :return: reduced dict, key is RioFeature, value is FeatureValue
        """
        if data.get('type', None) != 'FeatureCollection' or not isinstance(data.get('features', None), list):
            return dict()
        features_api = data.get('features', [])
        result = dict()
        for f in features_api:
            props = f.get('properties', {})
            if (f.get('id', None) is None or
                    props.get('value', None) is None):
                continue
            if (props.get('timestamp', None) is None and
                    props.get('date', None) is None):
                continue

            try:
                if 'timestamp' in props.keys():
                    timestamp = datetime.fromisoformat(props.get('timestamp'))
                else:
                    # Cut last character as the date is written '2024-06-15Z' which is not ISO compliant
                    timestamp = date.fromisoformat(props.get('date')[:-1])
                value = float(props.get('value'))
            except (TypeError, ValueError):
                continue

            name = f"{prefix}:{f.get('id').split('.')[0]}"
            if name not in [f'{f}' for f in features]:
                continue
            if name not in result or result[name]['timestamp'] < timestamp:
                result[name] = FeatureValue(timestamp=timestamp, value=value)

        return result


class IrcelineRioIfdmClient(IrcelineBaseWmsClient):
    """
    API client for RIO IFDM interpolated IRCEL - CELINE open data
    RIO IFDM is more fine-grained for interpolation than RIO but only allows one feature to be request at a time, which
    may be slower
    """
    _base_url = _rio_ifdm_wms_base_url

    async def get_data(self,
                       features: List[RioIfdmFeature],
                       position: Tuple[float, float]
                       ) -> Dict[RioIfdmFeature, FeatureValue]:
        """
        Get interpolated concentrations for the given features at the given position.
        :param features: pollutants to get the forecasts for
        :param position: (lat, long)
        :return: dict where key is RioIfdmFeature and value is a FeatureValue
        """
        result = dict()
        lat, lon = position
        base_querystring = (self._default_querystring |
                            {"bbox": f"{lon},{lat},{lon + self._epsilon},{lat + self._epsilon}"})

        tasks = [asyncio.create_task(self._get_single_feature(base_querystring, feature)) for feature in features]
        results = await asyncio.gather(*tasks)

        for r in results:
            result |= r

        return result

    async def _get_single_feature(self, base_querystring: dict, feature: RioIfdmFeature) -> dict:
        result = dict()
        querystring = base_querystring | {"layers": f"{feature}", "query_layers": f"{feature}"}
        try:
            r: ClientResponse = await self._api_wrapper(self._base_url, querystring)
            r: dict = await r.json()
            result[feature] = FeatureValue(
                value=r.get('features', [{}])[0].get('properties', {}).get('GRAY_INDEX'),
                timestamp=datetime.fromisoformat(r.get('timeStamp')) if 'timeStamp' in r else None)
        except (IrcelineApiError, ClientResponseError, IndexError):
            result[feature] = FeatureValue(value=None, timestamp=None)
        return result



from datetime import datetime, date
from enum import StrEnum
from typing import TypedDict


class IrcelineFeature(StrEnum):
    pass


class RioFeature(IrcelineFeature):
    BC_24HMEAN = 'rio:bc_24hmean'
    BC_DMEAN = 'rio:bc_dmean'
    BC_HMEAN = 'rio:bc_hmean'
    NO2_ANMEAN = 'rio:no2_anmean'
    NO2_DMEAN = 'rio:no2_dmean'
    NO2_HMEAN = 'rio:no2_hmean'
    O3_8HMEAN = 'rio:o3_8hmean'
    O3_ANMEAN = 'rio:o3_anmean'
    O3_HMEAN = 'rio:o3_hmean'
    O3_MAX8HMEAN = 'rio:o3_max8hmean'
    O3_MAXHMEAN = 'rio:o3_maxhmean'
    PM10_24HMEAN = 'rio:pm10_24hmean'
    PM10_ANMEAN = 'rio:pm10_anmean'
    PM10_DMEAN = 'rio:pm10_dmean'
    PM10_HMEAN = 'rio:pm10_hmean'
    PM25_24HMEAN = 'rio:pm25_24hmean'
    PM25_ANMEAN = 'rio:pm25_anmean'
    PM25_DMEAN = 'rio:pm25_dmean'
    PM25_HMEAN = 'rio:pm25_hmean'
    SO2_HMEAN = 'rio:so2_hmean'


class RioIfdmFeature(IrcelineFeature):
    PM25_HMEAN = 'rioifdm:pm25_hmean'
    NO2_HMEAN = 'rioifdm:no2_hmean'
    PM10_HMEAN = 'rioifdm:pm10_hmean'
    O3_HMEAN = 'rioifdm:o3_hmean'
    BELAQI = 'rioifdm:belaqi'


class ForecastFeature(IrcelineFeature):
    NO2_MAXHMEAN = 'forecast:no2_maxhmean'
    NO2_DMEAN = 'forecast:no2_dmean'
    O3_MAXHMEAN = 'forecast:o3_maxhmean'
    O3_MAX8HMEAN = 'forecast:o3_max8hmean'
    PM10_DMEAN = 'forecast:pm10_dmean'
    PM25_DMEAN = 'forecast:pm25_dmean'
    BELAQI = 'forecast:belaqi'


class FeatureValue(TypedDict):
    # Timestamp at which the value was computed
    timestamp: datetime | date | None
    value: int | float | None



from datetime import datetime, date

from freezegun import freeze_time

from src.open_irceline import IrcelineRioClient
from src.open_irceline.api import _rio_wfs_base_url, _user_agent
from src.open_irceline.data import RioFeature, FeatureValue
from src.open_irceline.utils import epsg_transform
from tests.conftest import get_api_data, get_mock_session


@freeze_time(datetime.fromisoformat("2024-06-15T16:55:03.419Z"))
async def test_format_result_hmean():
    data = get_api_data('rio_wfs.json')
    result = IrcelineRioClient._format_result('rio', data, [RioFeature.NO2_HMEAN, RioFeature.O3_HMEAN])

    expected = {
        str(RioFeature.O3_HMEAN): FeatureValue(
            timestamp=datetime.fromisoformat("2024-06-15T16:00:00Z"),
            value=71
        ),
        str(RioFeature.NO2_HMEAN): FeatureValue(
            timestamp=datetime.fromisoformat("2024-06-15T16:00:00Z"),
            value=4
        )
    }

    assert result == expected


@freeze_time(datetime.fromisoformat("2024-06-15T16:55:03.419Z"))
async def test_format_result_hmean_with_error():
    data = get_api_data('rio_wfs_with_errors.json')
    result = IrcelineRioClient._format_result('rio', data,
                                              [RioFeature.NO2_HMEAN, RioFeature.O3_HMEAN, RioFeature.PM25_HMEAN])

    expected = {
        str(RioFeature.O3_HMEAN): FeatureValue(
            timestamp=datetime.fromisoformat("2024-06-15T16:00:00Z"),
            value=71
        ),
        str(RioFeature.PM25_HMEAN): FeatureValue(
            timestamp=datetime.fromisoformat("2024-06-15T15:00:00Z"),
            value=1
        )
    }

    assert result == expected


@freeze_time(datetime.fromisoformat("2024-06-15T19:30:09.581Z"))
async def test_format_result_dmean():
    data = get_api_data('rio_wfs_dmean.json')
    result = IrcelineRioClient._format_result('rio', data,
                                              [RioFeature.BC_DMEAN, RioFeature.PM10_DMEAN, RioFeature.PM25_DMEAN])

    expected = {
        str(RioFeature.BC_DMEAN): FeatureValue(timestamp=date(2024, 6, 15), value=0.1),
        str(RioFeature.PM10_DMEAN): FeatureValue(timestamp=date(2024, 6, 15), value=5.9),
        str(RioFeature.PM25_DMEAN): FeatureValue(timestamp=date(2024, 6, 15), value=1.1),
    }

    assert result == expected


def test_parse_capabilities():
    data = get_api_data('rio_capabilities.xml', plain=True)
    result = IrcelineRioClient._parse_capabilities(data)

    expected = {'rio:so2_anmean_be', 'rio:o3_hmean', 'rio:bc_anmean_vl', 'rio:o3_anmean_be', 'rio:pm10_hmean_vl',
                'rio:o3_aot40for_be', 'rio:no2_maxhmean', 'rio:pm10_24hmean_1x1', 'rio:o3_aot40veg_5y_be',
                'rio:pm10_24hmean_vl', 'rio:pm10_anmean', 'rio:pm25_anmean_be', 'rio:o3_net60_be', 'rio:pm25_hmean_vl',
                'rio:so2_25thmax_hmean_be', 'rio:so2_hmean_vl', 'rio:pm10_dmean', 'rio:bc_hmean_wl',
                'rio:o3_aot40veg_be', 'rio:o3_max8hmean', 'rio:pm25_24hmean', 'rio:no2_hmean',
                'rio:rio_grid_4x4_centroids', 'rio:no2_anmean', 'rio:so2_hmean', 'rio:o3_hmean_vl', 'rio:rio_grid_4x4',
                'rio:pm10_hmean_1x1', 'rio:pm25_anmean_vl', 'rio:pm25_dmean', 'rio:rio_grid_1x1', 'rio:o3_aot60_be',
                'rio:no2_anmean_be', 'rio:o3_net60_3y_be', 'rio:pm10_anmean_be', 'rio:o3_maxhmean',
                'rio:no2_19thmax_hmean_be', 'rio:bc_anmean_be', 'rio:bc_24hmean', 'rio:no2_anmean_vl',
                'rio:no2_hmean_wl', 'rio:pm10_24hmean', 'rio:bc_hmean_vl', 'rio:pm10_anmean_vl', 'rio:pm10_hmean_wl',
                'rio:pm25_anmean', 'rio:o3_aot40for_5y_be', 'rio:rio_grid_4x4_be', 'rio:o3_8hmean',
                'rio:pm25_hmean_1x1', 'rio:no2_hmean_vl', 'rio:o3_anmean', 'rio:pm10_excday_be', 'rio:o3_anmean_vl',
                'rio:pm10_hmean', 'rio:pm25_24hmean_vl', 'rio:pm25_hmean', 'rio:bc_hmean', 'rio:so2_anmean_vl',
                'rio:bc_dmean'}

    assert result == expected


def test_parse_capabilities_with_error():
    result = IrcelineRioClient._parse_capabilities("wow there no valid XML")
    assert result == set()


async def test_api_rio():
    pos = (50.4657, 4.8647)
    x, y = epsg_transform(pos)
    session = get_mock_session('rio_wfs.json')

    client = IrcelineRioClient(session)

    d = date(2024, 6, 18)
    features = [RioFeature.NO2_HMEAN, RioFeature.O3_HMEAN]
    _ = await client.get_data(features, pos, d)
    session.request.assert_called_once_with(
        method='GET',
        url=_rio_wfs_base_url,
        params={"service": "WFS",
                "version": "1.3.0",
                "request": "GetFeature",
                "outputFormat": "application/json",
                "typeName": ",".join(features),
                "cql_filter":
                    f"date>='2024-06-17'"
                    f" AND "
                    f"INTERSECTS(the_geom, POINT ({x} {y}))"},
        headers={'User-Agent': _user_agent}
    )


async def test_api_rio_get_capabilities():
    session = get_mock_session(text_file='rio_capabilities.xml')

    client = IrcelineRioClient(session)
    _ = await client.get_capabilities()

    session.request.assert_called_once_with(
        method='GET',
        url=_rio_wfs_base_url,
        params={"service": "WFS",
                "version": "1.3.0",
                "request": "GetCapabilities"},
        headers={'User-Agent': _user_agent}
    )



from datetime import datetime
from unittest.mock import call

from freezegun import freeze_time

from src.open_irceline.api import _rio_ifdm_wms_base_url, _user_agent
from src.open_irceline.data import RioIfdmFeature, FeatureValue
from src.open_irceline.rio import IrcelineRioIfdmClient
from tests.conftest import get_api_data, get_mock_session


def test_parse_capabilities():
    data = get_api_data('rio_ifdm_capabilities.xml', plain=True)
    result = IrcelineRioIfdmClient._parse_capabilities(data)

    expected = {'no2_dm3', 'belaqi_dm2', 'pm10_hmean', 'belaqi_dm1', 'pm25_dm3', 'pm25_dm2', 'pm10_dm1', 'o3_dm3',
                'no2_dm1', 'pm10_dm3', 'pm25_dm1', 'belaqi', 'belaqi_dm3', 'pm10_dm2', 'o3_dm2', 'pm25_hmean', 'o3_dm1',
                'o3_hmean', 'no2_dm2', 'no2_hmean'}

    assert result == expected

    for f in RioIfdmFeature:
        assert f"{f.split(':')[1]}" in result


async def test_aget_capabilities():
    session = get_mock_session(text_file='rio_ifdm_capabilities.xml')

    client = IrcelineRioIfdmClient(session)
    _ = await client.get_capabilities()

    session.request.assert_called_once_with(
        method='GET',
        url=_rio_ifdm_wms_base_url,
        params={"service": "WMS",
                "version": "1.1.1",
                "request": "GetCapabilities"},
        headers={'User-Agent': _user_agent}
    )


@freeze_time(datetime.fromisoformat("2024-06-30T13:00:21.520Z"))
async def test_api_forecast_error():
    pos = (50.4657, 4.8647)
    session = get_mock_session('forecast_wms_feature_info_invalid.json')

    client = IrcelineRioIfdmClient(session)

    features = [RioIfdmFeature.NO2_HMEAN, RioIfdmFeature.O3_HMEAN]
    result = await client.get_data(features, pos)

    for k, v in result.items():
        assert v == FeatureValue(timestamp=datetime.fromisoformat("2024-06-30T13:00:21.520Z"), value=None)


async def test_api_forecast():
    pos = (50.4657, 4.8647)
    lat, lon = pos
    session = get_mock_session('forecast_wms_feature_info.json')

    client = IrcelineRioIfdmClient(session)

    features = [RioIfdmFeature.NO2_HMEAN, RioIfdmFeature.O3_HMEAN]
    result = await client.get_data(features, pos)

    base = {"service": "WMS",
            "version": "1.1.1",
            "request": "GetFeatureInfo",
            "info_format": "application/json",
            "width": "1",
            "height": "1",
            "srs": "EPSG:4326",
            "bbox": f"{lon},{lat},{lon + 0.00001},{lat + 0.00001}",
            "X": "1",
            "Y": "1"}

    calls = [call(
        method='GET',
        url=_rio_ifdm_wms_base_url,
        params=base | {"layers": f"{feature}",
                       "query_layers": f"{feature}"},
        headers={'User-Agent': _user_agent},
    )
        for feature in features]

    session.request.assert_has_calls(calls, any_order=True)

    for k, v in result.items():
        assert v['value'] == 10.853286743164062


async def test_api_forecast_no_field():
    pos = (50.4657, 4.8647)
    session = get_mock_session('forecast_wms_feature_info_no_field.json')

    client = IrcelineRioIfdmClient(session)

    features = [RioIfdmFeature.NO2_HMEAN, RioIfdmFeature.O3_HMEAN]
    result = await client.get_data(features, pos)

    for k, v in result.items():
        assert v == FeatureValue(timestamp=None, value=None)



from typing import Tuple

from pyproj import Transformer

_project_transform = Transformer.from_crs('EPSG:4326', 'EPSG:31370', always_xy=False)


def epsg_transform(position: Tuple[float, float]) -> Tuple[int, int]:
    """
    Convert 'EPSG:4326' coordinates to 'EPSG:31370' coordinates
    :param position: (x, y) coordinates
    :return: tuple of int in the EPSG:31370 system
    """
    result = _project_transform.transform(position[0], position[1])
    return round(result[0]), round(result[1])
